/*
 * VeriSilicon VPE Video Decoder Common interface
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdint.h>
#include <stdio.h>
#include <stddef.h>
#include "vpi_api.h"
#include "vpi_types.h"

#include "codec_internal.h"
#include "decode.h"
#include "libavutil/hwcontext_vpe.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/pixfmt.h"
#include "libavutil/rational.h"
#include "internal.h"
#include "hwconfig.h"
#if CONFIG_LIBXRM
#include "xrm_dec_interface.h"
#endif

#include "log_ama.h"

typedef struct VpeDecCtx {
    const AVClass *class;

#if CONFIG_LIBXRM
    XrmDecodeContext xrm_ctx;
#endif
    VpiDecInitParams *dec_params;
    VpiDecHandle dec_handle;

    int flag_decoder_err;

    // buffered packet feed to external decoder
    VpiPkt *buffered_pkt;
    void *pkt_pool;

    /* decoder options */
    int low_latency;
    char *format_str;
    char *extra_params;
    int input_depth;
    int output_depth; //The output bit depth requested from the decoder

    AVCodecParserContext *parser;
    int flag_av1_hwdev_not_init;

    char instance_name[64];
} VpeDecCtx;

#define PKT_BUF_POOL_SIZE 128

static int vpe_decode_get_avpixfmt_from_vpiformat(VpiFmt format)
{
    switch (format) {
    case VPI_FORMAT_NV12:
        return AV_PIX_FMT_NV12;
    case VPI_FORMAT_NV21:
        return AV_PIX_FMT_NV21;
    case VPI_FORMAT_P010LE:
        return AV_PIX_FMT_P010LE;
    case VPI_FORMAT_P010BE:
        return AV_PIX_FMT_P010BE;
    case VPI_FORMAT_YUV420P:
        return AV_PIX_FMT_YUV420P;
    case VPI_FORMAT_YUV420P10LE:
        return AV_PIX_FMT_YUV420P10LE;
    case VPI_FORMAT_YUV420P10BE:
        return AV_PIX_FMT_YUV420P10BE;
    case VPI_FORMAT_YUV422P:
        return AV_PIX_FMT_YUV422P;
    case VPI_FORMAT_YUV422P10LE:
        return AV_PIX_FMT_YUV422P10LE;
    case VPI_FORMAT_YUV422P10BE:
        return AV_PIX_FMT_YUV422P10BE;
    case VPI_FORMAT_YUV444P:
        return AV_PIX_FMT_YUV444P;
    case VPI_FORMAT_ARGB:
        return AV_PIX_FMT_ARGB;
    case VPI_FORMAT_ABGR:
        return AV_PIX_FMT_ABGR;
    case VPI_FORMAT_RGBA:
        return AV_PIX_FMT_RGBA;
    case VPI_FORMAT_BGRA:
        return AV_PIX_FMT_BGRA;
    case VPI_FORMAT_BGR0:
        return AV_PIX_FMT_BGR0;
    case VPI_FORMAT_RGB24:
        return AV_PIX_FMT_RGB24;
    case VPI_FORMAT_BGR24:
        return AV_PIX_FMT_BGR24;
    case VPI_FORMAT_RGB16:
        return AV_PIX_FMT_RGB565LE;
    case VPI_FORMAT_VPE:
        return AV_PIX_FMT_VPE;
    case VPI_FORMAT_RGB24_P:
        return AV_PIX_FMT_RGBP;
    case VPI_FORMAT_RGB48:
        return AV_PIX_FMT_R16G16B16;
    case VPI_FORMAT_RGB48_P:
        return AV_PIX_FMT_R16G16B16P;
    case VPI_FORMAT_BGR24_P:
        return AV_PIX_FMT_BGRP;
    case VPI_FORMAT_BGR48:
        return AV_PIX_FMT_B16G16R16;
    case VPI_FORMAT_BGR48_P:
        return AV_PIX_FMT_B16G16R16P;
    case VPI_FORMAT_SP1010:
        return AV_PIX_FMT_PACKED10;
    default:
        av_log(NULL, AV_LOG_ERROR, "unsupport VPI format %d\n", format);
        return -1;
    }
}

/**
 * Initialize hw frame and device
 *
 * The avctx->hw_frames_ctx is the reference to the AVHWFramesContext.
 * Create the hw_frame_ctx and initialize hw_frame_ctx.
 */
static int vpe_dec_init_hwctx(AVCodecContext *avctx)
{
    VpeDecCtx *dec_ctx = avctx->priv_data;
    AVHWFramesContext *hwframe_ctx;
    AVVpeHWConfig *hwconfig = NULL;
    AVHWFramesConstraints *constraints = NULL;
    int ret = 0;
    int i;
    enum AVPixelFormat sw_format;
    char *p;
    char *tmp = NULL;
    const AVPixFmtDescriptor *pix_desc;

    if (!avctx->hw_frames_ctx) {
        if (avctx->hw_device_ctx) {
            avctx->hw_frames_ctx = av_hwframe_ctx_alloc(avctx->hw_device_ctx);
            if (!avctx->hw_frames_ctx) {
                av_log(avctx, AV_LOG_ERROR, "av_hwframe_ctx_alloc failed\n");
                ret = AVERROR(ENOMEM);
                goto error;
            }
        } else {
            av_log(avctx, AV_LOG_ERROR, "No hw frame/device available\n");
            ret = AVERROR_UNKNOWN;
            goto error;
        }
    }


    /* for parsing much more packets */
    av_log(avctx, AV_LOG_DEBUG, "[%s@%d], parse width %d, height %d\n",  __FUNCTION__, __LINE__, avctx->width, avctx->height);
    hwconfig = av_hwdevice_hwconfig_alloc(avctx->hw_device_ctx);
    if (!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    switch (avctx->codec_id) {
    case AV_CODEC_ID_H264:
        hwconfig->id  = VPE_HW_H264DEC;
        break;
    case AV_CODEC_ID_HEVC:
        hwconfig->id  = VPE_HW_HEVCDEC;
        break;
    case AV_CODEC_ID_VP9:
        hwconfig->id = VPE_HW_VP9DEC;
        break;
    case AV_CODEC_ID_AV1:
        hwconfig->id = VPE_HW_AV1DEC;
        break;
    case AV_CODEC_ID_MJPEG:
        hwconfig->id = VPE_HW_JPEGDEC;
        break;
    default:
        ret = AVERROR(EINVAL);
        goto error;
    }
    av_log(avctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    constraints =
        av_hwdevice_get_hwframe_constraints(avctx->hw_device_ctx, hwconfig);
    if (!constraints) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    av_log(avctx, AV_LOG_DEBUG, "get constraints %d,%d,%d,%d\n",
           constraints->min_width, constraints->min_height,
           constraints->max_width, constraints->max_height);

    if (avctx->coded_width  < constraints->min_width  ||
        avctx->coded_height < constraints->min_height ||
        avctx->coded_width  > constraints->max_width  ||
        avctx->coded_height > constraints->max_height) {
        av_log(avctx, AV_LOG_ERROR, "Hardware does not support image "
                "size %dx%d (constraints: width %d-%d height %d-%d).\n",
                avctx->coded_width, avctx->coded_height,
                constraints->min_width,  constraints->max_width,
                constraints->min_height, constraints->max_height);
        ret = AVERROR(EINVAL);
        goto error;
    }

    if (!constraints->valid_sw_formats ||
        constraints->valid_sw_formats[0] == AV_PIX_FMT_NONE) {
        av_log(avctx, AV_LOG_ERROR, "Hardware does not offer any "
               "usable surface formats.\n");
        ret = AVERROR(EINVAL);
        goto error;
    }

    if (avctx->codec_id == AV_CODEC_ID_MJPEG) {
        if (avctx->pix_fmt == AV_PIX_FMT_YUVJ420P
                || avctx->pix_fmt == AV_PIX_FMT_YUVJ422P
                || avctx->pix_fmt == AV_PIX_FMT_YUVJ411P
                || avctx->pix_fmt == AV_PIX_FMT_YUVJ444P
                || avctx->pix_fmt == AV_PIX_FMT_YUVJ440P
                || avctx->pix_fmt == AV_PIX_FMT_GRAY8) {
            i = 0;
            while (1) {
                if (constraints->valid_sw_formats[i] == avctx->pix_fmt)
                    break;
                if (constraints->valid_sw_formats[i] == AV_PIX_FMT_NONE) {
                    av_log(avctx, AV_LOG_ERROR, "Hardware can't find any matched sw format\n");
                    ret = AVERROR(EINVAL);
                    goto error;
                }
                i++;
            }
        }
    }
    if (!constraints->valid_hw_formats ||
        constraints->valid_hw_formats[0] != AV_PIX_FMT_VPE) {
        av_log(avctx, AV_LOG_ERROR, "Hardware error format\n");
        ret = AVERROR(EINVAL);
        goto error;
    }
    av_log(avctx, AV_LOG_DEBUG, "match hw constraints %s\n", av_get_pix_fmt_name(constraints->valid_hw_formats[0]));

    pix_desc = av_pix_fmt_desc_get(avctx->pix_fmt);
    if(!pix_desc){
        av_log(avctx, AV_LOG_ERROR, "Bitstream not supported by this decoder, unable to get input pixel format\n");
        ret = AVERROR_INVALIDDATA;
        goto error;
    }
    //Figure out the input bit depth.
    dec_ctx->input_depth = pix_desc->comp[0].depth;
    av_log(avctx, AV_LOG_DEBUG, "input depth is %d\n", dec_ctx->input_depth);
    if(!(dec_ctx->input_depth == 8 || dec_ctx->input_depth ==10)){
        av_log(avctx, AV_LOG_ERROR, "Invalid bit depth(%d), only 8bit or 10bit input supported\n", dec_ctx->input_depth);
        ret =  AVERROR_INVALIDDATA;
        goto error;
    }

    //Parse the out_fmt string to figure out if the output sw_format
    // and the bit depth of the output
    dec_ctx->output_depth = 0;
    if(dec_ctx->format_str == NULL){
        //No output format specified we will use nv12-ma or p010le-ma
        if (dec_ctx->input_depth == 8){
            sw_format = AV_PIX_FMT_NV12;
            dec_ctx->output_depth = 8;
        } else if (dec_ctx->input_depth == 10){
            sw_format = AV_PIX_FMT_PACKED10; //CR-1155857 Change default 10-bit output to packed10
            dec_ctx->output_depth = 10;
        } else {
            av_log(avctx, AV_LOG_ERROR, "Only 8 and 10 bit input supported\n");
            ret = AVERROR_INVALIDDATA;
            goto error;
        }
    }else{
        tmp = (char*)malloc(strlen(dec_ctx->format_str) + 1);
        if(tmp == NULL){
            av_log(avctx, AV_LOG_ERROR, "Unable to allocate memory for tmp string\n");
            ret = AVERROR(ENOMEM);
            goto error;
        }

        strcpy(tmp, dec_ctx->format_str);
        p = strtok(tmp, "-");
        av_log(avctx, AV_LOG_INFO, "out_format = %s\n", p);
        if(p!=NULL){
            if(!strcmp(p, "nv12")){
                sw_format = AV_PIX_FMT_NV12;
                dec_ctx->output_depth = 8;
            }
            else if(!strcmp(p, "p010le")){
                sw_format = AV_PIX_FMT_P010LE;
                dec_ctx->output_depth = 10;
            }
            else if(!strcmp(p, "yuv420p")){
                sw_format = AV_PIX_FMT_YUV420P;
                dec_ctx->output_depth = 8;
            }
            else if(!strcmp(p, "yuv420p10le")){
                sw_format = AV_PIX_FMT_YUV420P10LE;
                dec_ctx->output_depth = 10;
            }
            else if (!strcmp(p, "packed10")){
                sw_format = AV_PIX_FMT_PACKED10;
                dec_ctx->output_depth = 10;
            }
            else if (!strcmp(p, "rgbp")){
                sw_format = AV_PIX_FMT_RGBP;
                dec_ctx->output_depth = 8; //Allow 10 bit inputs to be converted to rgbp ???
            }
            else{
                av_log(avctx, AV_LOG_ERROR, "Unsupported format(%s)\n", dec_ctx->format_str);
                ret = AVERROR(EINVAL);
                free(tmp);
                goto error;
            }
        }else{
            av_log(avctx, AV_LOG_ERROR, "Invalid format string(%s)\n", dec_ctx->format_str);
            ret = AVERROR(EINVAL);
            free(tmp);
            goto error;
        }
        free(tmp);
    }

    hwframe_ctx = (AVHWFramesContext *)avctx->hw_frames_ctx->data;
    hwframe_ctx->format = AV_PIX_FMT_VPE;
    hwframe_ctx->sw_format = sw_format;
    hwframe_ctx->width  = avctx->width;
    hwframe_ctx->height = avctx->height;

    av_log(avctx, AV_LOG_DEBUG, "%s(%d) avctx->pix_fmt %s (%d); hwframe_ctx->sw_format %s (%d)\n", __FUNCTION__, __LINE__,
            av_get_pix_fmt_name(avctx->pix_fmt),  avctx->pix_fmt,
            av_get_pix_fmt_name(hwframe_ctx->sw_format), hwframe_ctx->sw_format);

    if ((ret = av_hwframe_ctx_init(avctx->hw_frames_ctx)) < 0) {
        av_log(avctx, AV_LOG_ERROR, "av_hwframe_ctx_init failed\n");
        goto error;
    }

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return 0;

error:
    av_log(avctx, AV_LOG_DEBUG, "vpe_dec_init_hwctx failed\n");

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return ret;
}

/**
 * Get VPE error string
 */
static int vpe_decode_error(int err)
{
    av_log(NULL, AV_LOG_ERROR, "VPE decoder error: %s\n",
           vpi_error_str(err));
    return AVERROR_EXTERNAL;
}

/**
 * alloc frame from hwcontext for external codec
 */
static int vpe_decode_get_frame_buf(AVCodecContext *avctx, AVFrame *frame, VpiFrm **vpi_frame)
{
    int ret;

    ret = ff_get_buffer(avctx, frame, AV_GET_BUFFER_FLAG_REF);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "return ret %d\n", ret);
        *vpi_frame = NULL;
        return -1;
    }

    *vpi_frame = (VpiFrm *)frame->data[0];
    return 0;
}

static int vpe_dec_set_params(AVCodecContext *avctx)
{
    VpeDecCtx *dec_ctx = avctx->priv_data;
    AVVpeDeviceContext *device_hwctx;
    AVVpeFramesContext *frame_hwctx;
    AVHWDeviceContext *device_ctx;
    AVHWFramesContext *hwframe_ctx;
    VpiDecInitParams *dec_params;
    int input_depth, ret;

    hwframe_ctx  = (AVHWFramesContext *)avctx->hw_frames_ctx->data;
    frame_hwctx  = (AVVpeFramesContext *)hwframe_ctx->hwctx;
    device_ctx   = (AVHWDeviceContext *)avctx->hw_device_ctx->data;
    device_hwctx = (AVVpeDeviceContext *)device_ctx->hwctx;

    av_log(avctx, AV_LOG_DEBUG, "device_ctx = %p, hwframe_ctx = %p\n", device_ctx, hwframe_ctx);
    av_log(avctx, AV_LOG_DEBUG, "device_hwctx = %p\n", device_hwctx);
    av_log(avctx, AV_LOG_DEBUG, "device_hwctx->vpe_dev_ctx = %p\n", device_hwctx->vpe_dev_ctx);
    av_log(avctx, AV_LOG_DEBUG, "frame_hwctx->vpe_mem_ctx = %p\n", frame_hwctx->vpe_mem_ctx);

    ret = vpi_pool_create(frame_hwctx->vpe_mem_ctx, PKT_BUF_POOL_SIZE, &dec_ctx->pkt_pool);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "vpi create pkt pool failed, %d\n", vpe_decode_error(ret));
        return AVERROR(ENOMEM);
    }

    if(dec_ctx->output_depth != 0 && dec_ctx->input_depth != dec_ctx->output_depth){
        av_log(avctx, AV_LOG_ERROR, "Input bit depth(%d) and Output bit depth(%d) do not match\n", input_depth, dec_ctx->output_depth);
        return AVERROR_INVALIDDATA;
    }

    dec_ctx->dec_params = (VpiDecInitParams *)vpi_memory_get_struct(VPE_DEC_INIT_PARAM);
    dec_params = dec_ctx->dec_params;
    dec_params->dev_ctx = device_hwctx->vpe_dev_ctx;
    dec_params->mem_ctx = frame_hwctx->vpe_mem_ctx;
    dec_params->outfmt_str = dec_ctx->format_str;
    dec_params->src_width  = avctx->width;
    dec_params->src_height = avctx->height;
    dec_params->src_depth  = dec_ctx->input_depth;
    dec_params->pkt_pool = dec_ctx->pkt_pool;
    dec_params->low_latency = dec_ctx->low_latency;

    switch (avctx->codec_id) {
    case AV_CODEC_ID_H264:
        dec_params->dec_type = VPE_DEC_H264;
        break;
    case AV_CODEC_ID_HEVC:
        dec_params->dec_type = VPE_DEC_HEVC;
        break;
    case AV_CODEC_ID_VP9:
        dec_params->dec_type = VPE_DEC_VP9;
        break;
    case AV_CODEC_ID_AV1:
        dec_params->dec_type = VPE_DEC_AV1;
        break;
    case AV_CODEC_ID_MJPEG:
        dec_params->dec_type = VPE_DEC_JPEG;
        break;
    default:
        return -1;
    }

    return 0;
}

static void set_instance_name(AVCodecContext *avctx)
{
    VpeDecCtx *dec_ctx = avctx->priv_data;
    snprintf(dec_ctx->instance_name, 64, "%s@%p", (*(AVClass **)avctx)->item_name(avctx), avctx);
}

static av_cold int _vpe_decode_init(AVCodecContext *avctx)
{
    VpeDecCtx *dec_ctx = avctx->priv_data;
    int ret = 0;
#if CONFIG_LIBXRM
    XrmInterfaceProperties xrm_props;
    int dev_index = -1;
    memset(&xrm_props, 0, sizeof(xrm_props));
#endif
    ret = vpe_dec_init_hwctx(avctx);
    if (ret) {
        //FIX: av1 decoder not open hw device at avformat_find_stream_info, so sw parse av1 stream.
        if (AVERROR_UNKNOWN == ret && avctx->codec_id == AV_CODEC_ID_AV1) {
            av_log(avctx, AV_LOG_DEBUG, "av1 decoder hw dev not init\n");

            dec_ctx->parser = av_parser_init(AV_CODEC_ID_AV1);
            if (!dec_ctx->parser) {
                av_log(avctx, AV_LOG_DEBUG, "av1 parser not found\n");
                return AVERROR_DEMUXER_NOT_FOUND;
            }

            dec_ctx->flag_av1_hwdev_not_init = 1;
            return 0;
        } else {
            av_log(avctx, AV_LOG_ERROR, "dec init hwctx failed!\n");
            return -1;
        }
    }

    if (vpe_dec_set_params(avctx)) {
        av_log(avctx, AV_LOG_ERROR, "dec set params failed!\n");
        return AVERROR_UNKNOWN;
    }

#if CONFIG_LIBXRM
    // Reserve xrm resource
    xrm_props.width = avctx->width;
    xrm_props.height = avctx->height;
    xrm_props.fps_num = avctx->framerate.num;
    xrm_props.fps_den = avctx->framerate.den;

    if(xrm_props.fps_num == 0 || xrm_props.fps_den ==0){
        //This is possible if timing_info_present_flag is not set for bitstream
        //Default to calculating fps from pkt_timebase
        double fps_from_timebase = 1/(av_q2d(avctx->pkt_timebase));
        if(fps_from_timebase <= 0 || fps_from_timebase > 60){
            //Looks like pkt_timebase is not reliable defaulting to 30fps.
            av_log(avctx, AV_LOG_WARNING, "framerate info not available from bitstream defaulting to 30 fps\n");
            xrm_props.fps_num = 30;
            xrm_props.fps_den = 1;
        } else {
            av_log(avctx, AV_LOG_WARNING, "framerate info not available from bitstream using pkt_timebase (%f fps)\n", fps_from_timebase);
            xrm_props.fps_num = avctx->pkt_timebase.den;
            xrm_props.fps_den = avctx->pkt_timebase.num;
        }
    }

    if(vpi_device_get_device_index(dec_ctx->dec_params->dev_ctx, &dev_index) < 0) {
        return AVERROR(EINVAL);
    }

    if (xrm_dec_reserve(&dec_ctx->xrm_ctx, dev_index, &xrm_props) == XRM_ERROR) {
        return AVERROR(EBUSY);
    }
#endif

    dec_ctx->dec_params->extradata = NULL;
    dec_ctx->dec_params->extradata_size = 0;
    if (avctx->codec_id == AV_CODEC_ID_H264 || avctx->codec_id == AV_CODEC_ID_HEVC) {
        if (!avctx->internal->is_copy) {
            if (avctx->extradata_size > 0 && avctx->extradata) {
                uint8_t *data = avctx->extradata;
                if (avctx->codec_id == AV_CODEC_ID_H264 && data[0] == 1) {
                    av_log(avctx, AV_LOG_DEBUG, "avcC\n");
                } else if (avctx->codec_id == AV_CODEC_ID_HEVC &&
                      avctx->extradata_size > 3
                      && (data[0] || data[1] || data[2] > 1)) {
                    av_log(avctx, AV_LOG_DEBUG, "hvcC\n");
                } else {
                    av_log(avctx, AV_LOG_DEBUG, "avctx->extradata_size %d\n", avctx->extradata_size);
                    dec_ctx->dec_params->extradata = avctx->extradata;
                    dec_ctx->dec_params->extradata_size = avctx->extradata_size;
                }
            }
        }
    }

    dec_ctx->dec_handle = vpi_decoder_init(dec_ctx->dec_params);
    if (dec_ctx->dec_handle == NULL) {
        av_log(avctx, AV_LOG_ERROR, "decoder open failed!\n");
        return AVERROR_EXTERNAL;
    }


    if (dec_ctx->extra_params) {
        AVDictionary *dict    = NULL;
        AVDictionaryEntry *en = NULL;

        if (!av_dict_parse_string(&dict, dec_ctx->extra_params, "=", ":", 0)) {
            while ((en = av_dict_get(dict, "", en, AV_DICT_IGNORE_SUFFIX))) {
                if (vpi_decoder_set_param(dec_ctx->dec_handle, en->key, en->value) < 0)
                    av_log(avctx, AV_LOG_WARNING,
                           "Error parsing option '%s = %s'. Please check if it is valid for vpe dec extra options.\n",
                           en->key, en->value);
            }
            av_dict_free(&dict);
        }
    }

    ret = vpi_decoder_open(dec_ctx->dec_handle);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "vpi_processor_open failed, error=%d\n", vpe_decode_error(ret));
        return AVERROR_EXTERNAL;
    }

    av_log(avctx, AV_LOG_DEBUG, "vpe_decoder_open OK !!\n");

    dec_ctx->buffered_pkt = (VpiPkt *)vpi_memory_get_struct(VPE_PACKET);
    if (!dec_ctx->buffered_pkt) {
        return AVERROR(ENOMEM);
    }

    return 0;
}

static av_cold int vpe_decode_init(AVCodecContext *avctx)
{
    int ret;
    VpeDecCtx *dec_ctx = avctx->priv_data;
    set_instance_name(avctx);
    {
        LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_DEC, dec_ctx->instance_name);
        ret = _vpe_decode_init(avctx);
        LOG_AMA_RESTORE_INSTANCE_INFO();
    }
    return ret;
}

/**
 * Output the frame raw data
 */
static int vpe_decode_output_frame(AVCodecContext *avctx, AVFrame *frame, VpiFrm *pic)
{
    int i, ret = 0;
    AVHWFramesContext *hwframe_ctx = (AVHWFramesContext *)avctx->hw_frames_ctx->data;
    AVVpeFramesContext *frame_hwctx;
    int ch_index;

    frame_hwctx = (AVVpeFramesContext *)hwframe_ctx->hwctx;
    for (i = 0; i < VPE_CHANNEL_NUM; i++) {
        if (pic->ch_info[i].enabled == 1 &&
            pic->ch_info[i].sw_format != VPI_FORMAT_NONE) {
            frame_hwctx->sw_format[0] = vpe_decode_get_avpixfmt_from_vpiformat(pic->ch_info[i].sw_format);
            ch_index = i;
            break;
        }
    }
    if (ch_index == VPE_CHANNEL_NUM) {
        av_log(avctx, AV_LOG_ERROR, "No valid sw format\n");
        return AVERROR_EXTERNAL;
    }
    for (i = 0; i < VPE_CHANNEL_NUM; i++) {
        if (pic->ch_info[i].enabled == 1 &&
            pic->ch_info[i].sw_format != VPI_FORMAT_NONE &&
            i != ch_index) {
            if (frame_hwctx->sw_format[0] != vpe_decode_get_avpixfmt_from_vpiformat(pic->ch_info[i].sw_format)) {
                frame_hwctx->sw_format[1] = vpe_decode_get_avpixfmt_from_vpiformat(pic->ch_info[i].sw_format);
                break;
            }
        }
    }
    av_log(avctx, AV_LOG_DEBUG, "sw_format %d/%s, %d/%s\n",
                  frame_hwctx->sw_format[0], av_get_pix_fmt_name(frame_hwctx->sw_format[0]),
                  frame_hwctx->sw_format[1], av_get_pix_fmt_name(frame_hwctx->sw_format[1]));

    // check the decoder output frame width/height for resolution change stream
    if ((hwframe_ctx->width != pic->width) ||
        (hwframe_ctx->height != pic->height)) {
        av_log(avctx, AV_LOG_INFO, "resolution changed from %dx%d to %dx%d\n",
                                   hwframe_ctx->width, hwframe_ctx->height,
                                   pic->width, pic->height);
        hwframe_ctx->width = pic->width;
        hwframe_ctx->height = pic->height;
    }

    ret = ff_decode_frame_props(avctx, frame);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "ff_decode_frame_props failed\n");
        return ret;
    }

    frame->format      = AV_PIX_FMT_VPE;
    frame->width       = pic->width;
    frame->height      = pic->height;
    frame->linesize[0] = vpi_info_get_struct_size(VPE_FRAME);
    frame->pts         = pic->pts;
    frame->pkt_dts     = pic->dts;
    frame->key_frame   = pic->key_frame;
    frame->flags       = pic->flags & AV_FRAME_FLAG_DISCARD;

    return ret;
}

/**
 * receive VpiFrm from external decoder
 */
static int vpe_decode_get_frame(AVCodecContext *avctx, AVFrame *frame)
{
    VpeDecCtx *dec_ctx = (VpeDecCtx *)avctx->priv_data;
    VpiFrm *output_frm = NULL;
    AVBufferRef *ref = NULL;
    int ret, rv;

    ref = vpi_decoder_release_pkt(dec_ctx->dec_handle);
    while (ref) {
        /* unref the input avpkt buffer */
        av_buffer_unref(&ref);
        ref = vpi_decoder_release_pkt(dec_ctx->dec_handle);
    }

    ret = vpi_decoder_get_frame(dec_ctx->dec_handle);
    if (ret == VPI_DEC_FRAME) {
        if (vpe_decode_get_frame_buf(avctx, frame, &output_frm)) {
            return AVERROR(ENOMEM);
        }
        rv = vpi_decoder_get_output_frame(dec_ctx->dec_handle, output_frm);
        // get one output frame
        if (rv) {
            return vpe_decode_error(rv);
        } else {
            if (0 == vpe_decode_output_frame(avctx, frame, output_frm)) {
                if (frame->flags & AV_FRAME_FLAG_DISCARD) {
                    vpi_frame_release(output_frm);
                    av_frame_unref(frame);
                    return AVERROR(EAGAIN);
                } else {
                    return 0;
                }
            } else {
                return vpe_decode_error(ret);
            }
        }
    } else if (ret == VPI_DEC_EOF) {
        av_log(avctx, AV_LOG_DEBUG, "received EOF from external decoder\n");
        return AVERROR_EOF;
    } else if (ret == -1) {
        return vpe_decode_error(ret);
    } else {
        return AVERROR(EAGAIN);
    }
}

static int _vpe_decode_receive_frame(AVCodecContext *avctx, AVFrame *frame)
{
    VpeDecCtx *dec_ctx  = avctx->priv_data;
    AVBufferRef *ref    = NULL;
    AVPacket avpkt = { 0 };
    int ret = 0;
    int got_frame = 0;

    if (avctx->codec_id == AV_CODEC_ID_AV1 && dec_ctx->flag_av1_hwdev_not_init == 1) {
        ret = ff_decode_get_packet(avctx, &avpkt);
        av_log(avctx, AV_LOG_DEBUG, "ret %d, avpkt.data %p,  avpkt.size %d\n", ret, avpkt.data, avpkt.size);

        if (avpkt.size != 0 &&  avpkt.data != NULL){
            ret = av_parser_parse2(dec_ctx->parser, avctx, &avpkt.data,
                                   &avpkt.size, avpkt.data, avpkt.size,
                                   AV_NOPTS_VALUE, AV_NOPTS_VALUE, 0);
        }

        avctx->pix_fmt = dec_ctx->parser->format;
        frame->width   = dec_ctx->parser->width;
        frame->height  = dec_ctx->parser->height;
        frame->crop_top    = 0;
        frame->crop_bottom = 0;
        frame->crop_left   = 0;
        frame->crop_right  = 0;

        if (avpkt.size != 0) {
            av_packet_unref(&avpkt);
        }

        av_log(avctx, AV_LOG_DEBUG, "avctx->pix_fmt %d\n",  avctx->pix_fmt);

        if(!av_pix_fmt_desc_get(avctx->pix_fmt)){
            av_log(avctx, AV_LOG_ERROR, "Unable to get input pixel format\n");
            ret = AVERROR_INVALIDDATA;
            return ret;
        }

        if (!avctx->width || !avctx->height) {
            if ((int)frame->width != avctx->width || (int)frame->height != avctx->height) {
                av_log(avctx, AV_LOG_INFO, "av1 dimension change! %dx%d -> %dx%d\n",
                    avctx->width, avctx->height, frame->width, frame->height);
                ret = ff_set_dimensions(avctx, frame->width, frame->height);
                if (ret < 0)
                    return ret;
            }
        }

        return 0;
    }

    /* send the packet */
    ret = ff_decode_get_packet(avctx, &avpkt);
    av_log(avctx, AV_LOG_DEBUG, "[%s@%d],ret %d, avpkt.size %d, flags %d\n",
                                __FUNCTION__, __LINE__, ret, avpkt.size, avpkt.flags);
    if (ret == 0) {
        dec_ctx->buffered_pkt->buffer   = avpkt.data;
        dec_ctx->buffered_pkt->size     = avpkt.size;
        dec_ctx->buffered_pkt->pts      = avpkt.pts;
        dec_ctx->buffered_pkt->dts      = avpkt.dts;
        dec_ctx->buffered_pkt->duration = avpkt.duration;
        dec_ctx->buffered_pkt->flags    = avpkt.flags;
        ref                             = av_buffer_ref(avpkt.buf);
        if (!ref) {
            av_packet_unref(&avpkt);
            return AVERROR(ENOMEM);
        }
        dec_ctx->buffered_pkt->opaque = (void *)ref;
    } else if (ret == AVERROR_EOF) {
        av_log(avctx, AV_LOG_DEBUG, "get EOF packet\n");
        ret = vpi_decoder_put_packet(dec_ctx->dec_handle, dec_ctx->buffered_pkt);
        if (ret < 0) {
            return vpe_decode_error(ret);
        } else {
            ret = vpe_decode_get_frame(avctx, frame);
            return ret;
        }
    } else if (ret < 0) {
        return ret;
    }

retry:
    /* try to flush packet data */
    if (dec_ctx->buffered_pkt->size > 0) {
        ret = vpi_decoder_put_packet(dec_ctx->dec_handle, dec_ctx->buffered_pkt);
        if (ret > 0) {
            dec_ctx->buffered_pkt->size -= ret;
            if (dec_ctx->buffered_pkt->size != 0) {
                return vpe_decode_error(ret);
            }
            av_packet_unref(&avpkt);
        } else if (ret == 0) {
            // no empty pos to store packet and no dpb, check the decoder output
            // if has new output frame, dpb will be released
            if (0 == vpe_decode_get_frame(avctx, frame)) {
                got_frame = 1;
            }
            goto retry;
        } else {
            return vpe_decode_error(ret);
        }
    }

    if (got_frame != 1) {
        ret = vpe_decode_get_frame(avctx, frame);
    }
    return ret;
}

static int vpe_decode_receive_frame(AVCodecContext *avctx, AVFrame *frame)
{
    int ret;
    VpeDecCtx *dec_ctx = avctx->priv_data;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_DEC, dec_ctx->instance_name);
    ret = _vpe_decode_receive_frame(avctx, frame);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

static av_cold int _vpe_decode_close(AVCodecContext *avctx)
{
    VpeDecCtx *dec_ctx = avctx->priv_data;
    AVBufferRef *ref    = NULL;
#if CONFIG_LIBXRM
    xrm_dec_release(&dec_ctx->xrm_ctx);
#endif
    if (dec_ctx->dec_handle) {
        vpi_decoder_close(dec_ctx->dec_handle);
        if (dec_ctx->pkt_pool) {
            ref = vpi_decoder_flush_pkt(dec_ctx->pkt_pool);
            while (ref) {
                /* unref the input avpkt buffer */
                av_buffer_unref(&ref);
                ref = vpi_decoder_flush_pkt(dec_ctx->pkt_pool);
            }
            vpi_pool_unref(dec_ctx->pkt_pool);
        }
        dec_ctx->dec_handle = NULL;
    }

    if (dec_ctx->buffered_pkt) {
        vpi_memory_free_struct(dec_ctx->buffered_pkt);
    }

    if (dec_ctx->dec_params)
        vpi_memory_free_struct((void *)dec_ctx->dec_params);

    if (dec_ctx->flag_av1_hwdev_not_init == 1) {
        av_parser_close(dec_ctx->parser);
        dec_ctx->flag_av1_hwdev_not_init = 0;
    }

    return 0;
}

static av_cold int vpe_decode_close(AVCodecContext *avctx)
{
    int ret;
    VpeDecCtx *dec_ctx = avctx->priv_data;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_DEC, dec_ctx->instance_name);
    ret = _vpe_decode_close(avctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

#define OFFSET(x) offsetof(VpeDecCtx, x)
#define VD AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_DECODING_PARAM
const AVOption vpe_decode_options[] = {
    { "low_latency",
      "enable/disable lowlatency mode",
      OFFSET(low_latency),
      AV_OPT_TYPE_BOOL,
      { .i64 = 0 },
      0,
      1,
      VD },
    { "out_fmt",
      "set output format, available options = nv12,p010le,packed10,rgbp,yuv420p,yuv420p10le",
      OFFSET(format_str),
      AV_OPT_TYPE_STRING,
      {.str = NULL},
      0,
      0,
      VD},
    { "expert_options",
      "set dec extra params",
      OFFSET(extra_params),
      AV_OPT_TYPE_STRING,
      {.str=NULL},
      0,
      0,
      VD },
    { NULL },
};

static const AVCodecHWConfigInternal *vpe_hw_configs[] = {
    &(const AVCodecHWConfigInternal) {
        .public = {
            .pix_fmt     = AV_PIX_FMT_VPE,
            .methods     = AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX |
                           AV_CODEC_HW_CONFIG_METHOD_HW_FRAMES_CTX |
                           AV_CODEC_HW_CONFIG_METHOD_INTERNAL,
            .device_type = AV_HWDEVICE_TYPE_AMA,
        },
        .hwaccel = NULL,
    },
    NULL
};

static const AVClass vpe_decode_h264_class = {
    .class_name = "h264d_ama",
    .item_name  = av_default_item_name,
    .option     = vpe_decode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass vpe_decode_hevc_class = {
    .class_name = "hevcd_ama",
    .item_name  = av_default_item_name,
    .option     = vpe_decode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass vpe_decode_vp9_class = {
    .class_name = "vp9d_ama",
    .item_name  = av_default_item_name,
    .option     = vpe_decode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass vpe_decode_av1_class = {
    .class_name = "av1d_ama",
    .item_name  = av_default_item_name,
    .option     = vpe_decode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass vpe_decode_jpeg_class = {
    .class_name = "jpegd_ama",
    .item_name  = av_default_item_name,
    .option     = vpe_decode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

FFCodec ff_h264_vpe_decoder = {
    .p.name             = "h264_ama",
    .p.long_name        = NULL_IF_CONFIG_SMALL("H264 (VPE VC8000D)"),
    .p.type             = AVMEDIA_TYPE_VIDEO,
    .p.id               = AV_CODEC_ID_H264,
    .priv_data_size   = sizeof(VpeDecCtx),
    .init             = &vpe_decode_init,
    .close            = &vpe_decode_close,
    FF_CODEC_RECEIVE_FRAME_CB(vpe_decode_receive_frame),
    .p.priv_class       = &vpe_decode_h264_class,
    .p.capabilities     =  AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE | AV_CODEC_CAP_AVOID_PROBING,
    .p.pix_fmts         = (const enum AVPixelFormat[]) { AV_PIX_FMT_VPE,
                                                       AV_PIX_FMT_NONE },
  	.hw_configs       = vpe_hw_configs,
    .p.wrapper_name     = "ama",
    .bsfs             = "h264_mp4toannexb",
};

FFCodec ff_hevc_vpe_decoder = {
    .p.name             = "hevc_ama",
    .p.long_name        = NULL_IF_CONFIG_SMALL("HEVC (VPE VC8000D)"),
    .p.type             = AVMEDIA_TYPE_VIDEO,
    .p.id               = AV_CODEC_ID_HEVC,
    .priv_data_size   = sizeof(VpeDecCtx),
    .init             = vpe_decode_init,
    .close            = vpe_decode_close,
    FF_CODEC_RECEIVE_FRAME_CB(vpe_decode_receive_frame),
    .p.capabilities     = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE | AV_CODEC_CAP_AVOID_PROBING,
    .p.priv_class       = &vpe_decode_hevc_class,
    .p.pix_fmts         = (const enum AVPixelFormat[]) { AV_PIX_FMT_VPE,
                                                       AV_PIX_FMT_NONE },
    .hw_configs       = vpe_hw_configs,
    .p.wrapper_name     = "ama",
    .bsfs             = "hevc_mp4toannexb",
};

FFCodec ff_av1_vpe_decoder = {
    .p.name           = "av1_ama",
    .p.long_name      = NULL_IF_CONFIG_SMALL("AV1 (VPE VC8000D)"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_AV1,
    .priv_data_size = sizeof(VpeDecCtx),
    .init           = vpe_decode_init,
    .close          = vpe_decode_close,
    FF_CODEC_RECEIVE_FRAME_CB(vpe_decode_receive_frame),
    .p.capabilities   = AV_CODEC_CAP_DELAY| AV_CODEC_CAP_HARDWARE | AV_CODEC_CAP_AVOID_PROBING,
    .p.priv_class     = &vpe_decode_av1_class,
    .p.pix_fmts       = (const enum AVPixelFormat[]) { AV_PIX_FMT_VPE,
                                                     AV_PIX_FMT_NONE },
    .hw_configs     = vpe_hw_configs,
    .p.wrapper_name   = "ama",
};

FFCodec ff_jpeg_vpe_decoder = {
    .p.name           = "jpeg_ama",
    .p.long_name      = NULL_IF_CONFIG_SMALL("JPEG (VPE VC8000D)"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_MJPEG,
    .priv_data_size = sizeof(VpeDecCtx),
    .init           = vpe_decode_init,
    .close          = vpe_decode_close,
    FF_CODEC_RECEIVE_FRAME_CB(vpe_decode_receive_frame),
    .p.capabilities   = AV_CODEC_CAP_DELAY| AV_CODEC_CAP_HARDWARE | AV_CODEC_CAP_AVOID_PROBING,
    .p.priv_class     = &vpe_decode_jpeg_class,
    .p.pix_fmts       = (const enum AVPixelFormat[]) { AV_PIX_FMT_VPE,
                                                     AV_PIX_FMT_NONE },
    .hw_configs     = vpe_hw_configs,
    .p.wrapper_name   = "ama",
};
