/*
* Xilinx ROI side data overlay test filter
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include "vpi_types.h"

#include "avfilter.h"
#include "filters.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/hwcontext.h"

#define DUMP_ROI 0


//TBD: ROI implementation will eventually update as per VSI new side data framework XLNXS:129

typedef struct SBToOverlayContext {
    const AVClass *class;
    AVBufferRef       *hwframes_ref;
    int skip_frames_val;
    int highlight_threshold_val;
    int frame_cnt;
} SBToOverlayContext;


static int query_formats(AVFilterContext *avctx)
{
    AVFilterFormats  *infmts = NULL;
    AVFilterFormats *outfmts = NULL;
    const AVPixFmtDescriptor *desc;
    int err;

    for (desc = av_pix_fmt_desc_next(NULL); desc;
         desc = av_pix_fmt_desc_next(desc)) {
        if (desc->flags & AV_PIX_FMT_FLAG_HWACCEL)
            err = ff_add_format(&infmts,  av_pix_fmt_desc_get_id(desc));
        else
            err = ff_add_format(&outfmts, av_pix_fmt_desc_get_id(desc));
        if (err) {
            ff_formats_unref(&infmts);
            ff_formats_unref(&outfmts);
            return err;
        }
    }

    if ((err = ff_formats_ref(infmts,  &avctx->inputs[0]->outcfg.formats)) < 0 ||
        (err = ff_formats_ref(outfmts, &avctx->outputs[0]->incfg.formats)) < 0)
        return err;

    return 0;
}


static int config_inputs(AVFilterLink *inlink)
{
    AVFilterContext *avctx = inlink->dst;
    SBToOverlayContext *ctx = avctx->priv;

    av_buffer_unref(&ctx->hwframes_ref);

    ctx->frame_cnt = 0;

    if (!inlink->hw_frames_ctx) {
        av_log(ctx, AV_LOG_ERROR, "The input must have a hardware frame "
               "reference.\n");
        return AVERROR(EINVAL);
    }

    ctx->hwframes_ref = av_buffer_ref(inlink->hw_frames_ctx);
    if (!ctx->hwframes_ref)
        return AVERROR(ENOMEM);

    return 0;
}


static int config_output(AVFilterLink *outlink)
{
    AVFilterContext *avctx = outlink->src;
    AVFilterLink *inlink   = avctx->inputs[0];
    SBToOverlayContext *ctx = avctx->priv;
    enum AVPixelFormat *formats;
    int err, i, found;

    if (!ctx->hwframes_ref)
        return AVERROR(EINVAL);

    err = av_hwframe_transfer_get_formats(ctx->hwframes_ref,
                                          AV_HWFRAME_TRANSFER_DIRECTION_FROM,
                                          &formats, 0);
    if (err < 0)
        return err;

    found = 0;
    for (i = 0; formats[i] != AV_PIX_FMT_NONE; i++) {
        if (formats[i] == outlink->format) {
            found = 1;
            break;
        }
    }
    av_freep(&formats);

    if (!found) {
        av_log(ctx, AV_LOG_ERROR, "Invalid output format %s for hwframe "
               "download.\n", av_get_pix_fmt_name(outlink->format));
        return AVERROR(EINVAL);
    }

    outlink->w = inlink->w;
    outlink->h = inlink->h;

    return 0;
}


static int filter_frame(AVFilterLink *link, AVFrame *in_frame)
{
    AVFilterContext *ctx = link->dst;
    SBToOverlayContext *s = ctx->priv;
    VpiFrm *in1_pic = NULL;
    AVFrame *output = NULL;
    VpiSideband *sbinfo = NULL;
    unsigned char *detected_roi = NULL;
    VpiChannelInfo *roi_ch_info = NULL;
    int err;

    in1_pic = (VpiFrm *)in_frame->data[0];
    output = ff_get_video_buffer(ctx->outputs[0], in_frame->width,
                                 in_frame->height);
    if (!output) {
        err = AVERROR(ENOMEM);
    }
    err = av_hwframe_transfer_data(output, in_frame, 0);
    if (err < 0) {
        av_log(ctx, AV_LOG_ERROR, "sbtooverlay_test Filter Failed to download frame: %d.\n", err);
    }
    err = av_frame_copy_props(output, in_frame);

    sbinfo = (VpiSideband *)in1_pic->vpi_sideband[0];
    if (!sbinfo) {
        av_log(ctx, AV_LOG_ERROR, "sbtooverlay_test Filter can not find the sideband.\n");
        goto exit;
    }

    detected_roi = (unsigned char *)(sbinfo->rc_data);
    if (!detected_roi) {
        av_log(ctx, AV_LOG_ERROR, "sbtooverlay_test Filter can not find the ROI sideband data.\n");
        goto exit;
    }

    roi_ch_info = &(sbinfo->ch_info);

    s->frame_cnt++;

    // loop for the roi dimension
    for (int ry=0; ry<roi_ch_info->height; ry++) {
        for (int rx=0; rx<roi_ch_info->width; rx++) {
            int index = (ry*roi_ch_info->width) + rx;
            // if valid activation, loop for 16x16 box drawing
            if ((detected_roi[index] >= s->highlight_threshold_val)) {
                int locy=ry*16, locx=rx*16;
                for(int y = locy; y < locy+16; y++) {
                    for(int x = locx; x < locx+16; x++) {

                        // ensure to draw within frame dimension
                        if ((x < output->width) && (y < output->height)) {
                            int offsetY = 1 * ((x) + (y * output->width));

                            // use skip frame as per configuraion via filter parameter
                            if ((!s->skip_frames_val) || ((s->frame_cnt/s->skip_frames_val) & 0x01)) {
                                int temp = output->data[0][offsetY + 0] + ((detected_roi[index]*detected_roi[index])/(3*255));
                                if (temp > 255)
                                    temp = 255;
                                output->data[0][offsetY + 0] = temp;
                            }
                        }
                    }
                }
            }
        }
    }

#if DUMP_ROI
    FILE *fp = fopen("test_roi.dat","wb");
    if(fp)
    {
	int retVal;
        retVal = fwrite(detected_roi, 1, roi_ch_info->height*roi_ch_info->width, fp);
        if (!retVal)
            av_log(ctx, AV_LOG_ERROR, "sbtooverlay_test Filter Failed to write tensor dump returnVal: %d.\n", retVal);
        fclose(fp);
    }
    else
        av_log(ctx, AV_LOG_ERROR, "sbtooverlay_test Filter Failed with create/open file to write tensor dump.\n");
#endif

exit:
    av_frame_free(&in_frame);
    return ff_filter_frame(link->dst->outputs[0], output);
}


static av_cold void uninit(AVFilterContext *ctx)
{
    SBToOverlayContext *s = ctx->priv;
    av_buffer_unref(&s->hwframes_ref);
}


#define OFFSET(x) offsetof(SBToOverlayContext, x)
#define FLAGS AV_OPT_FLAG_FILTERING_PARAM | AV_OPT_FLAG_VIDEO_PARAM
static const AVOption sbtooverlay_test_options[] = {
    { "skip_frames",          "set the frames to skip for overlay highlight",   OFFSET(skip_frames_val), AV_OPT_TYPE_INT, {.i64=0}, 0, 255, FLAGS },
    { "highlight_threshold",  "set the threshold value for overlay highlight",  OFFSET(highlight_threshold_val), AV_OPT_TYPE_INT, {.i64=0}, 0, 255, FLAGS },
    { NULL }
};

AVFILTER_DEFINE_CLASS(sbtooverlay_test);

static const AVFilterPad avfilter_vf_sbtovoerlay_inputs[] = {
    {
        .name           = "default",
        .type           = AVMEDIA_TYPE_VIDEO,
        .filter_frame   = filter_frame,
        .config_props   = config_inputs,
    },
    { NULL }
};

static const AVFilterPad avfilter_vf_sbtovoerlay_outputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_VIDEO,
        .config_props = config_output,
    },
    { NULL }
};

AVFilter ff_vf_sbtooverlay_test = {
    .name              = "sbtooverlay_test",
    .description       = NULL_IF_CONFIG_SMALL("Download hwframe and highlight the roi detection on it."),
    .priv_size         = sizeof(SBToOverlayContext),
    .uninit            = uninit,
    .nb_inputs         = 1,
    .nb_outputs        = 1,
    FILTER_QUERY_FUNC(query_formats),
    .inputs            = avfilter_vf_sbtovoerlay_inputs,
    .outputs           = avfilter_vf_sbtovoerlay_outputs,
    .priv_class        = &sbtooverlay_test_class,
};
